import sys
import warnings
from typing import Dict, Tuple, Union
from GANDLF.losses import global_losses_dict
from GANDLF.metrics import global_metrics_dict
import torch
import torch.nn.functional as nnf

from GANDLF.utils import one_hot, reverse_one_hot, get_linear_interpolation_mode


def get_metric_output(
    metric_function: object,
    prediction: torch.Tensor,
    target: torch.Tensor,
    params: dict,
) -> Union[float, list]:
    """
    This function computes the metric output for a given metric function, prediction and target.

    Args:
        metric_function (object): The metric function to be used.
        prediction (torch.Tensor): The input prediction label for the corresponding image label.
        target (torch.Tensor): The input ground truth for the corresponding image label.
        params (dict): The parameters passed by the user yaml.

    Returns:
        float: The computed metric from the label and the prediction.
    """
    metric_output = metric_function(prediction, target, params).detach().cpu()

    if metric_output.dim() == 0:
        return metric_output.item()
    else:
        temp = metric_output.tolist()
        # this check is needed for precision
        if len(temp) > 1:
            return temp
        else:
            # TODO: this branch is extremely edge case and is buggy.
            #  Overall the case when metric returns a list but of length 1 is very rare. The only case is when
            #  the metric returns Nx.. tensor (i.e. without aggregation by elements) and batch_size==N==1. This branch
            #  would definitely fail for such a metrics like
            #  MulticlassAccuracy(num_classes=3, multidim_average="samplewise")
            #  Maybe the best solution is to raise an error here if metric is configured to return samplewise results?
            return metric_output.item()


def get_loss_and_metrics(
    image: torch.Tensor, target: torch.Tensor, prediction: torch.Tensor, params: dict
) -> Tuple[torch.Tensor, Dict[str, float]]:
    """
    This function computes the loss and metrics for a given image, ground truth and prediction output.

    Args:
        image (torch.Tensor): The input image stack according to requirements.
        target (torch.Tensor): The input ground truth for the corresponding image label.
        prediction (torch.Tensor): The input prediction label for the corresponding image label.
        params (dict): The parameters passed by the user yaml.

    Returns:
        Tuple[torch.Tensor, Dict[str,float]]: The computed loss and metrics from the label and the prediction.
    """
    # this is currently only happening for mse_torch
    if isinstance(params["loss_function"], dict):
        # check for mse_torch
        loss_function = global_losses_dict[list(params["loss_function"].keys())[0]]
    else:
        loss_str_lower = params["loss_function"].lower()
        assert (
            loss_str_lower in global_losses_dict
        ), f"Could not find the requested loss function '{params['loss_function']}'"
        loss_function = global_losses_dict[loss_str_lower]

    loss = 0
    # specialized loss function for sdnet
    sdnet_check = (len(prediction) > 1) and (params["model"]["architecture"] == "sdnet")

    if params["problem_type"] == "segmentation":
        target = one_hot(target, params["model"]["class_list"])

    deep_supervision_model = False
    if (
        (len(prediction) > 1)
        and not (sdnet_check)
        and ("deep" in params["model"]["architecture"])
    ):
        deep_supervision_model = True
        # this case is for models that have deep-supervision - currently only used for segmentation models
        # these weights are taken from previous publication (https://arxiv.org/pdf/2103.03759.pdf)
        loss_weights = [0.5, 0.25, 0.175, 0.075]

        assert len(prediction) == len(
            loss_weights
        ), "Loss weights must be same length as number of outputs."

        ground_truth_resampled = []
        ground_truth_prev = target.detach()
        for i, _ in enumerate(prediction):
            if ground_truth_prev[0].shape != prediction[i][0].shape:
                # we get the expected shape of resampled ground truth
                expected_shape = reverse_one_hot(
                    prediction[i][0].detach(), params["model"]["class_list"]
                ).shape

                # linear interpolation is needed because we want "soft" images for resampled ground truth
                ground_truth_prev = nnf.interpolate(
                    ground_truth_prev,
                    size=expected_shape,
                    mode=get_linear_interpolation_mode(len(expected_shape)),
                    align_corners=False,
                )
            ground_truth_resampled.append(ground_truth_prev)

    if sdnet_check:
        # this is specific for sdnet-style archs
        loss_seg = loss_function(prediction[0], target.squeeze(-1), params)
        loss_reco = global_losses_dict["l1"](prediction[1], image[:, :1, ...], None)
        loss_kld = global_losses_dict["kld"](prediction[2], prediction[3])
        loss_cycle = global_losses_dict["mse"](prediction[2], prediction[4], None)
        loss = 0.01 * loss_kld + loss_reco + 10 * loss_seg + loss_cycle
    elif deep_supervision_model:
        # this is for models that have deep-supervision
        for i, _ in enumerate(prediction):
            # loss is calculated based on resampled "soft" labels using a pre-defined weights array
            loss += (
                loss_function(prediction[i], ground_truth_resampled[i], params)
                * loss_weights[i]
            )
    else:
        loss = loss_function(prediction, target, params)
    metric_output = {}

    # Metrics should be a list
    for metric in params["metrics"]:
        metric_lower = metric.lower()
        metric_output[metric] = 0
        if metric_lower not in global_metrics_dict:
            warnings.warn("WARNING: Could not find the requested metric '" + metric)
            continue

        metric_function = global_metrics_dict[metric_lower]
        if sdnet_check:
            metric_output[metric] = get_metric_output(
                metric_function, prediction[0], target.squeeze(-1), params
            )
        elif deep_supervision_model:
            for i, _ in enumerate(prediction):
                metric_output[metric] += get_metric_output(
                    metric_function, prediction[i], ground_truth_resampled[i], params
                )
        else:
            metric_output[metric] = get_metric_output(
                metric_function, prediction, target, params
            )
    return loss, metric_output



import math
from typing import Callable, Iterable, List, Optional, Tuple

import torch
from torch import Tensor
from torch.optim import Optimizer


class AdEMAMix(Optimizer):
    r"""Adapted from https://github.com/frgfm/Holocron/blob/main/holocron/optim/ademamix.py
    
    Implements the AdEMAMix optimizer from `"The AdEMAMix Optimizer: Better, Faster, Older" <https://arxiv.org/pdf/2409.03137>`_.

    The estimation of momentums is described as follows, :math:`\forall t \geq 1`:

    .. math::
        m_{1,t} \leftarrow \beta_1 m_{1, t-1} + (1 - \beta_1) g_t \\
        m_{2,t} \leftarrow \beta_3 m_{2, t-1} + (1 - \beta_3) g_t \\
        s_t \leftarrow \beta_2 s_{t-1} + (1 - \beta_2) (g_t - m_t)^2 + \epsilon

    where :math:`g_t` is the gradient of :math:`\theta_t`,
    :math:`\beta_1, \beta_2, \beta_3 \in [0, 1]^3` are the exponential average smoothing coefficients,
    :math:`m_{1,0} = 0,\ m_{2,0} = 0,\ s_0 = 0`, :math:`\epsilon > 0`.

    Then we correct their biases using:

    .. math::
        \hat{m_{1,t}} \leftarrow \frac{m_{1,t}}{1 - \beta_1^t} \\
        \hat{s_t} \leftarrow \frac{s_t}{1 - \beta_2^t}

    And finally the update step is performed using the following rule:

    .. math::
        \theta_t \leftarrow \theta_{t-1} - \eta \frac{\hat{m_{1,t}} + \alpha m_{2,t}}{\sqrt{\hat{s_t}} + \epsilon}

    where :math:`\theta_t` is the parameter value at step :math:`t` (:math:`\theta_0` being the initialization value),
    :math:`\eta` is the learning rate, :math:`\alpha > 0` :math:`\epsilon > 0`.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining parameter groups
        lr (float, optional): learning rate
        betas (Tuple[float, float, float], optional): coefficients used for running averages (default: (0.9, 0.999, 0.9999))
        alpha (float, optional): the exponential decay rate of the second moment estimates (default: 5.0)
        eps (float, optional): term added to the denominator to improve numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (bool, optional): whether to use the AMSGrad variant (default: False)
    """

    def __init__(
        self,
        params: Iterable[torch.nn.Parameter],
        lr: float = 1e-3,
        betas: Tuple[float, float, float] = (0.9, 0.999, 0.9999),
        alpha: float = 5.0,
        eps: float = 1e-8,
        weight_decay: float = 0.0,
    ) -> None:
        assert lr >= 0.0, f"Invalid learning rate: {lr}"
        assert eps >= 0.0, f"Invalid epsilon value: {eps}"
        assert all(
            0.0 <= beta < 1.0 for beta in betas
        ), f"Invalid beta parameters: {betas}"
        defaults = {
            "lr": lr,
            "betas": betas,
            "alpha": alpha,
            "eps": eps,
            "weight_decay": weight_decay,
        }
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:  # type: ignore[override]
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            exp_avgs_slow = []
            exp_avg_sqs = []
            state_steps = []

            for p in group["params"]:
                if p.grad is not None:
                    params_with_grad.append(p)
                    if p.grad.is_sparse:
                        raise RuntimeError(
                            f"{self.__class__.__name__} does not support sparse gradients"
                        )
                    grads.append(p.grad)

                    state = self.state[p]
                    # Lazy state initialization
                    if len(state) == 0:
                        state["step"] = 0
                        # Exponential moving average of gradient values
                        state["exp_avg"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )
                        state["exp_avg_slow"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )
                        # Exponential moving average of squared gradient values
                        state["exp_avg_sq"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )

                    exp_avgs.append(state["exp_avg"])
                    exp_avgs_slow.append(state["exp_avg_slow"])
                    exp_avg_sqs.append(state["exp_avg_sq"])

                    # update the steps for each param group update
                    state["step"] += 1
                    # record the step after step update
                    state_steps.append(state["step"])

            beta1, beta2, beta3 = group["betas"]
            _update_ademamix(
                params_with_grad,
                grads,
                exp_avgs,
                exp_avgs_slow,
                exp_avg_sqs,
                state_steps,
                beta1,
                beta2,
                beta3,
                group["alpha"],
                group["lr"],
                group["weight_decay"],
                group["eps"],
            )
        return loss


def _update_ademamix(
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    exp_avgs_slow: List[Tensor],
    exp_avg_sqs: List[Tensor],
    state_steps: List[int],
    beta1: float,
    beta2: float,
    beta3: float,
    alpha: float,
    lr: float,
    weight_decay: float,
    eps: float,
) -> None:
    r"""Functional API that performs AdaBelief algorithm computation.
    See :class:`~holocron.optim.AdaBelief` for details.
    """
    for i, param in enumerate(params):
        grad = grads[i]
        m1 = exp_avgs[i]
        m2 = exp_avgs_slow[i]
        nu = exp_avg_sqs[i]
        step = state_steps[i]

        bias_correction1 = 1 - beta1**step
        bias_correction2 = 1 - beta2**step

        if weight_decay != 0:
            grad = grad.add(param, alpha=weight_decay)

        # Decay the first and second moment running average coefficient
        m1.mul_(beta1).add_(grad, alpha=1 - beta1)
        nu.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
        m2.mul_(beta3).add_(grad, alpha=1 - beta3)

        denom = (nu.sqrt() / math.sqrt(bias_correction2)).add_(eps)

        param.addcdiv_(m1 / bias_correction1 + alpha * m2, denom, value=-lr)


def ademamix_wrapper(parameters: dict) -> torch.optim.Optimizer:
    """
    Creates an AdEMAMix optimizer from the PyTorch `torch.optim` module using the input parameters.

    Args:
        parameters (dict): A dictionary containing the input parameters for the optimizer.

    Returns:
        torch.optim.Optimizer: An AdEMAMix optimizer.
    """

    return AdEMAMix(
        params=parameters["model_parameters"],
        lr=parameters.get("learning_rate", 1e-3),
        betas=parameters.get("betas", (0.9, 0.999, 0.9999)),
        alpha=parameters.get("alpha", 5.0),
        eps=parameters.get("eps", 1e-8),
        weight_decay=parameters.get("weight_decay", 0.0),
    )



from typing import Tuple
import torch
import torchio


def handle_nonempty_batch(subject: dict, params: dict) -> Tuple[dict, int]:
    """
    Function to detect batch size from the subject an Opacus loader provides in the case of a non-empty batch, and make any changes to the subject dictionary that are needed for GaNDLF to use it.

    Args:
        subject (dict): Training data subject dictionary.
        params (dict): Training parameters.

    Returns:
        Tuple[dict, int]: Modified subject dictionary and batch size.
    """
    batch_size = len(subject[params["channel_keys"][0]][torchio.DATA])
    return subject, batch_size


def handle_empty_batch(subject: dict, params: dict, feature_shape: list) -> dict:
    """
    Function to replace the list of empty arrays an Opacus loader provides in the case of an empty batch with a subject dictionary GANDLF can consume.

    Args:
        subject (dict): Training data subject dictionary.
        params (dict): Training parameters.
        feature_shape (list): Shape of the features.

    Returns:
        dict: Modified subject dictionary.
    """

    print("\nConstructing empty batch dictionary.\n")

    subject = {
        "subject_id": "empty_batch",
        "spacing": None,
        "path_to_metadata": None,
        "location": None,
    }
    subject.update(
        {
            key: {torchio.DATA: torch.zeros(tuple([0] + feature_shape))}
            for key in params["channel_keys"]
        }
    )
    if params["problem_type"] != "segmentation":
        subject.update(
            {
                key: torch.zeros((0, params["model"]["num_classes"])).to(torch.int64)
                for key in params["value_keys"]
            }
        )
    else:
        subject.update(
            {
                "label": {
                    torchio.DATA: torch.zeros(tuple([0] + feature_shape)).to(
                        torch.int64
                    )
                }
            }
        )

    return subject


def handle_dynamic_batch_size(subject: dict, params: dict) -> Tuple[dict, int]:
    """
    Function to process the subject Opacus loaders provide and prepare to handle their dynamic batch size (including possible empty batches).

    Args:
        subject (dict): Training data subject dictionary.
        params (dict): Training parameters.

    Raises:
        RuntimeError: If the subject is a list object that is not an empty batch.

    Returns:
        Tuple[dict, int]: Modified subject dictionary and batch size.
    """

    # The handling performed here is currently to be able to comprehend what
    # batch size we are currently working with (which we may later see as not needed)
    # and also to handle the previously observed case where Opacus produces
    # a subject that is not a dictionary but rather a list of empty arrays
    # (due to the empty batch result). The latter case is detected as a subject that
    # is a list object.
    if isinstance(subject, list):
        are_empty = torch.Tensor(
            [torch.equal(tensor, torch.Tensor([])) for tensor in subject]
        )
        assert torch.all(
            are_empty
        ), "Detected a list subject that is not an empty batch, which is not expected behavior."
        # feature_shape = [params["model"]["num_channels"]]+params["patch_size"]
        feature_shape = [params["model"]["num_channels"]] + params["patch_size"]
        subject = handle_empty_batch(
            subject=subject, params=params, feature_shape=feature_shape
        )
        batch_size = 0
    else:
        subject, batch_size = handle_nonempty_batch(subject=subject, params=params)

    return subject, batch_size
